<?php

namespace Drupal\widget_ingestion\Form;

use Drupal\Core\Form\ConfirmFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\widget_ingestion\WidgetRemoteRepository;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form class to ingest the widgets now.
 */
final class ManualIngestionForm extends ConfirmFormBase {

  /**
   * The repository of widgets.
   *
   * @var \Drupal\widget_ingestion\WidgetRemoteRepository
   */
  private $widgetsRepository;

  /**
   * WidgetIngestionCommands constructor.
   *
   * @param \Drupal\widget_ingestion\WidgetRemoteRepository $widgets_repository
   *   The repository.
   */
  public function __construct(WidgetRemoteRepository $widgets_repository) {
    $this->widgetsRepository = $widgets_repository;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static($container->get(WidgetRemoteRepository::class));
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'widget_ingestion_manual';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildForm($form, $form_state);
    $form['force_ingestion'] = [
      '#title' => $this->t('Force ingestion'),
      '#description' => $this->t('If checked, all widget definitions will be ingested. If unchecked, only the newer widget definitions will be ingested.'),
      '#type' => 'checkbox',
      '#default_value' => FALSE,
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Ingest the widgets now.
    $force_ingestion = (bool) $form_state->getValue('force_ingestion');
    $items = $this->widgetsRepository->listAll();
    $batch = [
      'title' => $this->t('Ingesting widget definitions'),
      'operations' => array_map(function (array $item) use ($force_ingestion) {
        return [
          __CLASS__ . '::ingestWidget',
          [$item, $force_ingestion],
        ];
      }, $items),
      'finished' => __CLASS__ . '::batchFinished',
    ];
    batch_set($batch);
  }

  /**
   * {@inheritdoc}
   */
  public static function batchFinished($success, $results, $operations, $elapsed_time): void {
    \Drupal::messenger()
      ->addMessage(t('Ingested all widgets in @time: %widgets.', [
        '@time' => $elapsed_time,
        '%widgets' => implode(', ', $results),
      ]));
  }

  /**
   * Processes one batch operation.
   *
   * @param array $data
   *   The widget data.
   * @param bool $force_ingestion
   *   Weather or not to force the ingestion.
   * @param mixed $context
   *   Operates the UI for the batch process.
   */
  public static function ingestWidget(array $data, bool $force_ingestion, &$context) {
    $name = $data['title'];
    $context['message'] = t('Ingesting widget type "@name".', ['@name' => $name]);
    \Drupal::service(WidgetRemoteRepository::class)
      ->ingestWidget($data['shortcode'], $force_ingestion);
    $context['results'][] = $name;
  }

  /**
   * {@inheritdoc}
   */
  public function getQuestion() {
    return $this->t('Do you want to ingest the widgets from the registry?');
  }

  /**
   * {@inheritdoc}
   *
   * @SuppressWarnings(PHPMD.StaticAccess)
   */
  public function getCancelUrl() {
    return Url::fromRoute('system.admin_content');
  }

  /**
   * {@inheritdoc}
   */
  public function getConfirmText() {
    return $this->t('Ingest');
  }

  /**
   * {@inheritdoc}
   */
  public function getDescription() {
    $endpoint = \Drupal::config('widget_ingestion.settings')->get('endpoint');
    return $this->t(
      'Manual ingestion will update the widget type entities with the latest information from the widget registry.<br /><br />Ingestion URL: %endpoint',
      ['%endpoint' => $endpoint]
    );
  }

}
