<?php

namespace Drupal\widget_ingestion;

use Drupal\Component\Assertion\Inspector;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\File\Exception\FileException;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Queue\QueueWorkerManagerInterface;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\GuzzleException;
use const DIRECTORY_SEPARATOR;
use const DRUPAL_ROOT;

/**
 * Service to interact with the widget registry.
 */
final class WidgetRemoteRepository {

  public const FORCE_INGESTION_KEY = '_force_ingestion';

  /**
   * The HTTP client.
   *
   * @var \GuzzleHttp\Client
   */
  private $client;

  /**
   * The widget registry endpoint.
   *
   * @var string
   */
  private $endpoint = '';

  /**
   * The queue worker.
   *
   * @var \Drupal\Core\Queue\QueueWorkerInterface
   */
  private $queueWorker;

  /**
   * The data from the widget registry.
   *
   * @var array|null
   */
  private $registryData;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  private $fileSystem;

  /**
   * WidgetRemoteRepository constructor.
   *
   * @param \GuzzleHttp\ClientInterface $client
   *   The guzzle client to make HTTP requests.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The configuration factory.
   * @param \Drupal\Core\Queue\QueueWorkerManagerInterface $queue_worker_manager
   *   The queue worker manager.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system service.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function __construct(ClientInterface $client, ConfigFactoryInterface $config_factory, QueueWorkerManagerInterface $queue_worker_manager, FileSystemInterface $file_system) {
    $this->client = $client;
    $settings = $config_factory->get('widget_ingestion.settings');
    $this->endpoint = $settings->get('endpoint');
    $this->queueWorker = $queue_worker_manager
      ->createInstance('widget_ingestion_queue');
    $this->fileSystem = $file_system;
  }

  /**
   * Gets a list of all widgets from the widget registry.
   *
   * @return array
   *   The list of all widgets.
   *
   * @SuppressWarnings(PHPMD.StaticAccess)
   */
  public function listAll(): array {
    if (isset($this->registryData)) {
      return $this->registryData;
    }
    try {
      $response = $this->client->get($this->endpoint, [
        'headers' => [
          'Accept' => 'application/json',
        ],
      ]);
      $this->registryData = Json::decode($response->getBody());
    }
    catch (GuzzleException $e) {
      watchdog_exception('widget_ingestion', $e);
      $this->registryData = [];
    }
    return $this->registryData;
  }

  /**
   * Re-ingest all the widgets.
   *
   * @param bool $force_ingestion
   *   Force ingesting the widget types regardless of the version comparison.
   *
   * @return array[]
   *   The queue items.
   */
  public function ingestAll(bool $force_ingestion = FALSE): array {
    $items = array_map(static function (array $item) use ($force_ingestion) {
      return array_merge($item, [self::FORCE_INGESTION_KEY => $force_ingestion]);
    }, $this->listAll());
    array_map([$this->queueWorker, 'processItem'], $items);
    return $items;
  }

  /**
   * Re-ingest a widget.
   *
   * @param string $short_code
   *   The widget's short code.
   * @param bool $force_ingestion
   *   Force ingesting the widget types regardless of the version comparison.
   *
   * @return array|null
   *   The queued item or NULL if not found.
   */
  public function ingestWidget(string $short_code, bool $force_ingestion = FALSE): ?array {
    $all_widgets = $this->listAll();
    $res = array_filter($all_widgets, static function (array $item) use ($short_code) {
      return $item['shortcode'] === $short_code;
    });
    $widget = reset($res);
    if (!$widget) {
      return NULL;
    }
    $widget[self::FORCE_INGESTION_KEY] = $force_ingestion;
    try {
      $this->queueWorker->processItem($widget);
    }
    catch (\Exception $exception) {
      watchdog_exception('widget_ingestion', $exception);
      return NULL;
    }
    return $widget;
  }

  /**
   * Ingests the assets for the widget type, if necessary.
   *
   * @param string $remote_directory
   *   The remote directory.
   * @param array $files
   *   The list of files to, potentially, ingest.
   * @param string $remote_id
   *   The machine name of the widget definition.
   * @param string $version
   *   The version of the widget.
   *
   * @return \Drupal\widget_ingestion\LocalAssetDescriptor
   *   The asset descriptor to build the Drupal library definition.
   */
  public function ingestAssets(string $remote_directory, array $files, string $remote_id, string $version): LocalAssetDescriptor {
    // All files should be strings.
    assert(Inspector::assertAllStrings($files));
    // We will treat the major versions as different folders.
    $major_version = preg_replace('/(v?\d+)\.\d+\.\d+.*/', '$1', $version);
    $local_dir_wrapper = sprintf('public://widget-data/%s/%s', $remote_id, $major_version);
    $temp_dir = $this->fileSystem->tempnam($this->fileSystem->getTempDirectory(), $remote_id . '-backup-');
    $has_backup = TRUE;
    try {
      $this->fileSystem->move($local_dir_wrapper, $temp_dir);
    }
    catch (FileException $exception) {
      $has_backup = FALSE;
    }
    try {
      array_map(function (string $file) use ($local_dir_wrapper, $remote_directory) {
        $local_file_wrapper = sprintf('%s/%s', $local_dir_wrapper, $file);
        // Ensure that all the intermediary directories are created.
        $dirname = $this->fileSystem->dirname($local_file_wrapper);
        if (!file_exists($dirname)) {
          $this->fileSystem->mkdir($dirname, NULL, TRUE);
        }
        $local_path = $this->fileSystem->realpath($local_file_wrapper);
        $remote_url = sprintf('%s/%s', $remote_directory, $file);
        $this->client->request('GET', $remote_url, ['sink' => $local_path]);
        return $local_path;
      }, $files);
    }
    catch (GuzzleException $exception) {
      // If there is a backup, put it back.
      if ($has_backup) {
        $this->fileSystem->move($temp_dir, $local_dir_wrapper);
      }
    }
    // Now calculate the relative dir path to the widget_type module.
    return new LocalAssetDescriptor(
      $this->relativePath($this->fileSystem->realpath($local_dir_wrapper)),
      $files
    );
  }

  /**
   * Takes a path and a base dir and it makes a relative path.
   *
   * @param string $path
   *   The input path.
   * @param string|null $base_dir
   *   The base directory. If not provided, it defaults to the app root.
   *
   * @return string
   *   The relative path.
   */
  private function relativePath(string $path, string $base_dir = NULL): string {
    if (empty($base_dir)) {
      $base_dir = \Drupal::getContainer()->hasParameter('app.root')
        ? \Drupal::getContainer()->getParameter('app.root')
        : DRUPAL_ROOT;
    }
    $path = preg_replace('/' . preg_quote($base_dir, '/') . '/', '', $path);
    $path = ltrim($path, DIRECTORY_SEPARATOR);
    $levels = count(explode(DIRECTORY_SEPARATOR, drupal_get_path('module', 'widget_type')));
    for ($level = 0; $level < $levels; $level++) {
      $path = '..' . DIRECTORY_SEPARATOR . $path;
    }
    return $path;
  }

}
